from collections.abc import Mapping, Sequence

from yankee.util import camelize


def do_nothing(obj):
    return obj


class JsonPath(object):
    def __init__(self, key, many=False):
        self.key = key.split(".")
        self.many = many

    def __call__(self, obj):
        try:
            for k in self.key:
                if isinstance(obj, Mapping):
                    obj = obj[k]
                elif isinstance(obj, Sequence) and k.isdigit():
                    obj = obj[int(k)]
                elif isinstance(obj, Sequence):
                    obj = [o[k] for o in obj]
                else:
                    raise IndexError(f"Cannot get key {k} from {obj}")
            if not self.many and isinstance(obj, list):
                raise ValueError(
                    f"Expected {'.'.join(self.key)} to produce single value, got many: {obj}"
                )
            elif self.many and not isinstance(obj, list):
                raise ValueError(
                    f"Expected {'.'.join(self.key)} to produce many values, got one: {obj}"
                )
            return obj
        except (IndexError, KeyError):
            if self.many:
                return list()
            return None


class JsonMixin(object):
    def make_accessor(self):
        if self.data_key == False:
            return do_nothing
        data_key = self.data_key or self.name
        if data_key is None:
            return do_nothing
        return JsonPath(camelize(data_key), many=self.many)



import datetime

import pytest

from yankee import Schema, fields as f

doc1 = {
    "string": "Some String Data",
    "date_time": "2021-05-04T12:05",
    "date": "2021-05-04",
    "booleans": ["True", "true", "False", "false"],
    "float": 1.234,
    "int": 23,
    "exists": "Something",
    "name": {"part1": "George", "part2": "Burdell"},
    "random": "Some data",
    "address_line_1": "1234 Anywhere",
    "address_line_2": "Austin, TX 71234",
    "bad_string": "",
}

def to_obj(obj):
    if isinstance(obj, dict):
        return type("Dict", (object,), {k: to_obj(v) for k, v in obj.items()})
    elif isinstance(obj, list):
        return [to_obj(i) for i in obj]
    else:
        return obj

doc2 = to_obj(doc1)

class AddressField(f.Combine):
    address_line_1 = f.Str()
    address_line_2 = f.Str()

    def combine_func(self, obj):
        return f"{obj.address_line_1}\n{obj.address_line_2}"

class NameSchema(f.Combine):
    part1 = f.Str()
    part2 = f.Str()

    def combine_func(self, obj):
        return f"{obj['part1']} {obj['part2']}"

class SubSchema(Schema):
    string = f.Str()
    float = f.Float()

class ExampleSchema(Schema):
    string = f.Str()
    date_time = f.DT()
    date = f.Date()
    booleans = f.List(f.Bool, data_key="booleans")
    float = f.Float()
    int = f.Int()
    exists = f.Exists()
    does_not_exist = f.Exists()
    name = NameSchema()
    sub = SubSchema(False)
    address = AddressField(False)
    bad_string = f.Str()


def test_fields_on_dict():
    schema = ExampleSchema()
    data = schema.load(doc1)
    assert data["string"] == "Some String Data"
    assert data["date_time"] == datetime.datetime(2021, 5, 4, 12, 5)
    assert data["date"] == datetime.date(2021, 5, 4)
    assert data["booleans"] == [True, True, False, False]
    assert data["float"] - 1.234 < 0.001
    assert data["int"] == 23
    assert data["exists"] == True
    assert data["does_not_exist"] == False
    assert data["name"] == "George Burdell"
    assert data['sub']['string'] == "Some String Data"
    assert data['address'] == "1234 Anywhere\nAustin, TX 71234"
    assert "bad_string" not in data

def test_fields_on_obj():
    schema = ExampleSchema()
    data = schema.load(doc2)
    assert data["string"] == "Some String Data"
    assert data["date_time"] == datetime.datetime(2021, 5, 4, 12, 5)
    assert data["date"] == datetime.date(2021, 5, 4)
    assert data["booleans"] == [True, True, False, False]
    assert data["float"] - 1.234 < 0.001
    assert data["int"] == 23
    assert data["exists"] == True
    assert data["does_not_exist"] == False
    assert data["name"] == "George Burdell"
    assert data['sub']['string'] == "Some String Data"
    assert data['address'] == "1234 Anywhere\nAustin, TX 71234"
    assert "bad_string" not in data


from yankee.base import fields

from .key import JsonMixin


class Field(JsonMixin, fields.Field):
    pass


class String(JsonMixin, fields.String):
    pass


class DateTime(JsonMixin, fields.DateTime):
    pass


class Date(JsonMixin, fields.Date):
    pass


class Boolean(JsonMixin, fields.Boolean):
    pass


class Float(JsonMixin, fields.Float):
    pass


class Integer(JsonMixin, fields.Integer):
    pass


class Exists(JsonMixin, fields.Exists):
    pass


class Const(JsonMixin, fields.Const):
    pass

class List(JsonMixin, fields.List):
    pass

class Dict(JsonMixin, fields.Dict):
    pass

class Combine(JsonMixin, fields.Combine):
    pass

class Alternative(JsonMixin, fields.Alternative):
    pass

class ZipSchema(JsonMixin, fields.ZipSchema):
    _list_field = List

class DelimitedString(JsonMixin, fields.DelimitedString):
    pass


# Aliases
Str = String
DT = DateTime
Bool = Boolean
Int = Integer
Alt = Alternative



import re

from yankee.util import camelize, underscore, is_valid, AttrDict

from .deserializer import Deserializer



def inflect(string, style=None):
    try:
        if style is None:
            return string
        elif style == "json":
            return camelize(string)
        elif style == "python":
            return underscore(string)
    except Exception:
        return None

class Schema(Deserializer):
    class Meta:
        output_style = "python"

    def __init__(
        self,
        data_key=None,
        flatten=False,
        prefix=False,
        required=False,
    ):
        self.flatten = flatten
        self.prefix = prefix
        super().__init__(data_key, required)

    def bind(self, name=None, parent=None):
        super().bind(name, parent)
        # Make sure that fields are grabbed from superclasses as well
        class_fields = list()
        for c in reversed(self.__class__.mro()):
            class_fields += [
                (k, v) for k, v in c.__dict__.items() if isinstance(v, Deserializer)
            ]
        fields = dict(class_fields)
        for name, field in fields.items():
            field.bind(name, self)
        self.fields = {self.get_output_name(k): v for k, v in fields.items()}

    def get_output_name(self, name):
        output_style = getattr(self.Meta, "output_style", None)
        if output_style == None:
            return name
        elif output_style == "json":
            name = camelize(name)
            if self.prefix:
                return camelize(self.name) + name[0].upper() + name[1:]
            return name
        elif output_style == "python":
            name = underscore(name)
            if self.prefix:
                return underscore(self.name) + "_" + name
            return name

    def deserialize(self, obj) -> "Dict":
        output = AttrDict()
        for key, field in self.fields.items():
            value = field.load(obj)
            # If there is no value, don't include anything in the output dictionary
            if not is_valid(value):
                if field.required == True:
                    return dict()
                continue
            # If the value isn't a dict, or there's not flatten directive, add and continue
            if not isinstance(value, dict) or not getattr(field, "flatten", False):
                output[key] = value
                continue
            # Merge in flattened fields
            output.update(value)
        return output


class PolymorphicSchema(Schema):
    def bind(self, name=None):
        super().bind(self, name)
        for schema in self.schemas:
            schema.bind(name)

    def choose_schema(self, obj):
        raise NotImplementedError("Must be implemented in subclass!")

    def deserialize(self, raw_obj) -> "Dict":
        # Get the key one time only, rather than
        # on both deserializing the selector obj
        # and the final output
        obj = super().deserialize(obj)
        schema = self.choose_schema(obj)
        return schema.deserialize(raw_obj)

class RegexSchema(Schema):
    """
    This schema type allows for using a regex to pull data
    out of a string, and then treat it like a schema
    """
    __regex__ = None
    
    def __init__(self, *args, **kwargs):
        self._regex = re.compile(self.__regex__)
        super().__init__(*args, **kwargs)
        
    def deserialize(self, obj):
        if obj is None:
            return None
        text = self.to_string(obj)
        match = self._regex.search(text)
        if match is None:
            return None
        data = self.convert_groupdict(match.groupdict())
        return super().deserialize(data)
    
    def convert_groupdict(self, obj):
        return obj


import re

us_re_1 = re.compile(r"([A-Z]+)([A-Z][a-z])")
us_re_2 = re.compile(r"([a-z\d])([A-Z])")
us_re_3 = re.compile(r"([^\d])(\d+)")


def underscore(word: str) -> str:
    """
    Modified version from inflection library
    that also underscores digits
    """
    word = us_re_1.sub(r"\1_\2", word)
    word = us_re_2.sub(r"\1_\2", word)
    word = us_re_3.sub(r"\1_\2", word)
    word = word.replace("-", "_")
    return word.lower()


camelize_re = re.compile(r"(?:^|_)(.)")


def camelize(string: str) -> str:
    """
    Optimized version of the camelize function in inflections
    """
    result = camelize_re.sub(lambda m: m.group(1).upper(), string)
    return result[0].lower() + result[1:]


def is_valid(obj):
    if obj is None:
        return False
    elif isinstance(obj, (int, float)):
        return True
    elif isinstance(obj, (dict, list, str)) and len(obj) == 0:
        return False
    return True


# Cleans whitespace from text data
whitespace_re = re.compile(r"\s+")
clean_whitespace = lambda s: whitespace_re.sub(" ", s).strip().strip(",").strip()

def do_nothing(obj):
    return obj

class AttrDict(dict):    
    def __getattr__(self,  name):
        return self[name]
    
    def __setattr__(self, name, value):
        self[name] = value
    
    @classmethod
    def convert(cls, obj):
        if isinstance(obj, dict):
            return cls((k, cls.convert(v)) for k, v in obj.items())
        elif isinstance(obj, list):
            return list(cls.convert(i) for i in obj)
        else:
            return obj


from yankee.base import schema
from yankee.base.deserializer import Deserializer

from .key import XmlMixin

class Deserializer(XmlMixin, Deserializer):
    pass

class Schema(XmlMixin, schema.Schema):
    pass


class PolymorphicSchema(XmlMixin, schema.PolymorphicSchema):
    pass


class RegexSchema(XmlMixin, schema.RegexSchema):
    pass